/*
 * Copyright (C) 2007 OpenedHand Ltd
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <config.h>
#include <glib/gi18n.h>
#include <libecal/e-cal-view.h>

#include "koto-group-store.h"

G_DEFINE_TYPE (KotoGroupStore, koto_group_store, GTK_TYPE_LIST_STORE);

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), KOTO_TYPE_GROUP_STORE, KotoGroupStorePrivate))

typedef struct
{
  KotoGroupStoreType type;
  ECalView *view;
  /** A hash of group names to GtkTreeIter* */
  GHashTable *iter_hash;
  guint sig_added, sig_modified;
} KotoGroupStorePrivate;

enum {
  PROP_NULL,
  PROP_TYPE,
  PROP_CALVIEW,
};

/*
 * Private methods.
 */

static void
on_objects_added (ECalView *view, GList *objects, KotoGroupStore *store)
{
  KotoGroupStorePrivate *priv = GET_PRIVATE (store);

  for (;objects; objects = g_list_next (objects)) {
    icalcomponent *comp = objects->data;
    icalproperty *prop;
    char **categories, **l;
    
    prop = icalcomponent_get_first_property (comp, ICAL_CATEGORIES_PROPERTY);
    if (!prop)
      continue;

    categories = g_strsplit (icalproperty_get_categories (prop), ",", 0);

    for (l = categories; *l; l++) {
      GtkTreeIter iter;
      
      if (!g_hash_table_lookup (priv->iter_hash, *l)) {
        gtk_list_store_insert_with_values (GTK_LIST_STORE (store),
                                           &iter, 0,
                                           COL_TYPE, GROUP_TYPE_CATEGORY,
                                           COL_NAME, *l,
                                           -1);
        g_hash_table_insert (priv->iter_hash, g_strdup (*l), gtk_tree_iter_copy (&iter));
      }
    }

    g_strfreev (categories);
    icalproperty_free (prop);
  }
}

static gint
sorter_cb (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer user_data)
{
  int type_a, type_b;
  int res;
  char *name_a, *name_b;

  gtk_tree_model_get (model, a, COL_TYPE, &type_a, -1);
  gtk_tree_model_get (model, b, COL_TYPE, &type_b, -1);

  if (type_a != type_b)
    return type_a - type_b;

  gtk_tree_model_get (model, a, COL_NAME, &name_a, -1);
  gtk_tree_model_get (model, b, COL_NAME, &name_b, -1);

  res = g_utf8_collate (name_a ?: "", name_b ?: "");

  g_free (name_a);
  g_free (name_b);
  
  return res;
}


/*
 * Object methods.
 */

static void
koto_group_store_get_property (GObject *object, guint property_id,
                              GValue *value, GParamSpec *pspec)
{
  KotoGroupStorePrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_TYPE:
    g_value_set_int (value, priv->type);
    break;
  case PROP_CALVIEW:
    g_value_set_object (value, priv->view);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
koto_group_store_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
  KotoGroupStorePrivate *priv = GET_PRIVATE (object);
  /* GTK+ 2.8 won't let us pass a NULL iterator */
  GtkTreeIter iter;

  switch (property_id) {
  case PROP_TYPE:
    priv->type = g_value_get_int (value);
    switch (priv->type) {
    case KOTO_GROUP_STORE_TYPE_FILTER:
      gtk_list_store_insert_with_values (GTK_LIST_STORE (object), &iter, 0,
                                         COL_TYPE, GROUP_TYPE_ALL,
                                         COL_NAME, _("All"),
                                         -1);
      gtk_list_store_insert_with_values (GTK_LIST_STORE (object), &iter, 1,
                                         COL_TYPE, GROUP_TYPE_SEPARATOR, -1);
      break;
    case KOTO_GROUP_STORE_TYPE_SELECTOR:
      gtk_list_store_insert_with_values (GTK_LIST_STORE (object), &iter, 0,
                                         COL_TYPE, GROUP_TYPE_NONE,
                                         COL_NAME, _("None"),
                                         -1);
      gtk_list_store_insert_with_values (GTK_LIST_STORE (object), &iter, 1,
                                         COL_TYPE, GROUP_TYPE_SEPARATOR, -1);
      gtk_list_store_insert_with_values (GTK_LIST_STORE (object), &iter, 0,
                                         COL_TYPE, GROUP_TYPE_NEW,
                                         COL_NAME, _("New Group..."),
                                         -1);
      break;
    default:
      g_warning ("Unhandled KotoGroupStoreType %d", priv->type);
    }
  break;
  case PROP_CALVIEW:
    koto_group_store_set_view (KOTO_GROUP_STORE (object),
                               g_value_get_object (value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
koto_group_store_dispose (GObject *object)
{
  koto_group_store_set_view (KOTO_GROUP_STORE (object), NULL);

  if (G_OBJECT_CLASS (koto_group_store_parent_class)->dispose)
    G_OBJECT_CLASS (koto_group_store_parent_class)->dispose (object);
}

static void
koto_group_store_finalize (GObject *object)
{
  KotoGroupStorePrivate *priv = GET_PRIVATE (object);

  g_hash_table_destroy (priv->iter_hash);
  
  G_OBJECT_CLASS (koto_group_store_parent_class)->finalize (object);
}

static void
koto_group_store_class_init (KotoGroupStoreClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (KotoGroupStorePrivate));

  object_class->get_property = koto_group_store_get_property;
  object_class->set_property = koto_group_store_set_property;
  object_class->dispose = koto_group_store_dispose;
  object_class->finalize = koto_group_store_finalize;

  g_object_class_install_property (object_class, PROP_TYPE,
                                   g_param_spec_int ("type", "type", NULL,
                                                     0, G_MAXINT, KOTO_GROUP_STORE_TYPE_FILTER,
                                                     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
                                                     G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));

  g_object_class_install_property (object_class, PROP_CALVIEW,
                                   g_param_spec_object ("calview", "calview", NULL,
                                                        E_TYPE_CAL_VIEW,
                                                        G_PARAM_READWRITE |
                                                        G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));
}

static void
koto_group_store_init (KotoGroupStore *self)
{
  const GType column_types[] = { G_TYPE_INT, G_TYPE_STRING };
  KotoGroupStorePrivate *priv = GET_PRIVATE (self);

  gtk_list_store_set_column_types (GTK_LIST_STORE (self),
                                   G_N_ELEMENTS (column_types),
                                   (GType *) column_types);

  gtk_tree_sortable_set_default_sort_func (GTK_TREE_SORTABLE (self),
                                           sorter_cb, NULL, NULL);
  gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (self),
                                        GTK_TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID,
                                        GTK_SORT_ASCENDING);
  
  priv->iter_hash = g_hash_table_new_full (g_str_hash, g_str_equal,
                                      g_free, (GDestroyNotify)gtk_tree_iter_free);
}


/*
 * Public methods.
 */

GtkTreeModel *
koto_group_store_new (KotoGroupStoreType type, ECalView *view)
{
  return g_object_new (KOTO_TYPE_GROUP_STORE,
                       "type", type,
                       "calview", view,
                       NULL);
}

void
koto_group_store_set_view (KotoGroupStore *store, ECalView *view)
{
  KotoGroupStorePrivate *priv;

  g_return_if_fail (KOTO_IS_GROUP_STORE (store));

  priv = GET_PRIVATE (store);

  if (priv->view) {
    g_signal_handler_disconnect (priv->view, priv->sig_added);
    g_signal_handler_disconnect (priv->view, priv->sig_modified);
    priv->sig_added = priv->sig_modified = 0;
    g_object_unref (priv->view);
    priv->view = NULL;
  }
  
  if (view) {
    priv->view = g_object_ref (view);
    priv->sig_added = g_signal_connect (priv->view, "objects_added", G_CALLBACK (on_objects_added), store);
    priv->sig_modified = g_signal_connect (priv->view, "objects_modified", G_CALLBACK (on_objects_added), store);
  }
}

gboolean
koto_group_store_get_iter_for_group (KotoGroupStore *store, const char *group, GtkTreeIter *iter)
{
  KotoGroupStorePrivate *priv;
  GtkTreeIter *it;

  g_return_val_if_fail (KOTO_IS_GROUP_STORE (store), FALSE);

  priv = GET_PRIVATE (store);

  /* Handle empty groups by returning the first row (All or None) */
  if (group == NULL || group[0] == '\0') {
    return gtk_tree_model_get_iter_first (GTK_TREE_MODEL (store), iter);
  }
  
  if (g_hash_table_lookup_extended (priv->iter_hash, group, NULL, (gpointer*)(void*)&it)) {
    *iter = *it;
    return TRUE;
  } else {
    return FALSE;
  }
}

/**
 * koto_group_store_add_new_group:
 * @store: A #KotoGroupStore.
 * @iter: An optional #GtkTreeIter, which will be set to point at the new group.
 * @name: The name of the group to create.
 *
 * Create a new group called @name and set @iter to point at it.  If the group
 * already exists, @iter will point at the existing group.
 */
void
koto_group_store_add_new_group (KotoGroupStore *store, GtkTreeIter *iter, const char *name)
{
  KotoGroupStorePrivate *priv;
  GtkTreeIter *it;
  g_return_if_fail (KOTO_IS_GROUP_STORE (store));
  g_return_if_fail (name);
  
  priv = GET_PRIVATE (store);
  
  it = g_hash_table_lookup (priv->iter_hash, name);
  if (it) {
    *iter = *it;
  } else {
    gtk_list_store_insert_with_values (GTK_LIST_STORE (store),
                                       iter, 0,
                                       COL_TYPE, GROUP_TYPE_CATEGORY,
                                       COL_NAME, name,
                                       -1);
    g_hash_table_insert (priv->iter_hash, g_strdup (name), gtk_tree_iter_copy (iter));
  }
}
