/*
 * Copyright (C) 2007 OpenedHand Ltd
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <libical/ical.h>

#ifdef HAVE_SEXY
#include <libsexy/sexy-icon-entry.h>
#endif

#include "ical-util.h"
#include "koto-task.h"
#include "koto-date-combo.h"
#include "koto-group-combo.h"
#include "koto-group-store.h"
#include "koto-priority-combo.h"
#include "koto-task-editor.h"

G_DEFINE_TYPE (KotoTaskEditor, koto_task_editor, GTK_TYPE_TABLE);

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), KOTO_TYPE_TASK_EDITOR, KotoTaskEditorPrivate))

typedef struct
{
  KotoGroupStore *group_store;
  KotoTask *task;
  gboolean dirty;
  /* TODO: not sure if I really need this */
  gboolean ignore;
  GtkWidget **field_widgets;
} KotoTaskEditorPrivate;

enum {
  PROP_0,
  PROP_GROUPS,
  PROP_TASK,
  PROP_DIRTY,
  PROP_LAST,
};

#define FIELD_DATA "KotoTaskEditor:Field"
#define PROP_DATA "KotoTaskEditor:Property"
#define WIDGET_DATA "KotoTaskEditor:Widget"

/* Forward declaration for function pointer typedef */
typedef struct _DisplayField DisplayField;

/* Function that creates a widget for a particular field */
typedef GtkWidget* (*EditorNew) (KotoTaskEditor *editor);
/* Function that sets the widget value for a particular field/property */
typedef void (*EditorSetFunc) (GtkWidget *editor, icalproperty *prop);

/*
 * Struct defining a field to display.
 */
struct _DisplayField {
  char *property_name;
  icalproperty_kind kind;
  char *display_name;
  EditorNew editor_new;
  EditorSetFunc set_value;
};

static GtkWidget *entry_new (KotoTaskEditor *editor);
static void entry_set_from_string (GtkWidget *widget, icalproperty *prop);
static GtkWidget *url_entry_new (KotoTaskEditor *editor);
static GtkWidget *text_new (KotoTaskEditor *editor);
static void text_set_from_string (GtkWidget *widget, icalproperty *prop);
static GtkWidget *priority_new (KotoTaskEditor *editor);
static void priority_set (GtkWidget *widget, icalproperty *prop);
static GtkWidget *date_new (KotoTaskEditor *editor);
static void date_set (GtkWidget *widget, icalproperty *prop);
static GtkWidget *category_new (KotoTaskEditor *editor);
static void category_set (GtkWidget *widget, icalproperty *prop);

/*
 * List of fields to display in the main editor.  The order of this list is the
 * order they will be shown in.
 */
static const DisplayField fields[] = {
  { "show-summary", ICAL_SUMMARY_PROPERTY, N_("_Summary:"), entry_new, entry_set_from_string },
  { "show-category", ICAL_CATEGORIES_PROPERTY, N_("Ca_tegory:"), category_new, category_set },
  { "show-priority", ICAL_PRIORITY_PROPERTY, N_("_Priority:"), priority_new, priority_set },
  { "show-due-date", ICAL_DUE_PROPERTY, N_("D_ue Date:"), date_new, date_set },
  { "show-web-address", ICAL_URL_PROPERTY, N_("_Web Address:"), url_entry_new, entry_set_from_string },
  { "show-description", ICAL_DESCRIPTION_PROPERTY, N_("_Description:"), text_new, text_set_from_string },
};

/*
 * Callback when a GtkEntry field changes.
 */
static void
on_entry_changed (GtkEntry *entry, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  DisplayField *field;
  icalproperty *prop;
  const char *text;

  if (priv->ignore)
    return;

  field = g_object_get_data (G_OBJECT (entry), FIELD_DATA);
  prop = g_object_get_data (G_OBJECT (entry), PROP_DATA);

  if (!field) {
    g_warning (FIELD_DATA " property not set");
    return;
  }

  if (!prop) {
    prop = icalproperty_new (field->kind);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (entry), PROP_DATA, prop);
  }

  text = gtk_entry_get_text (entry);

  /* NO means use the default value type. Ugly as sin. */
  icalproperty_set_value_from_string (prop, text, "NO");

  priv->dirty = TRUE;
  return;
}

/*
 * Editor creation function (an EditorNew) , which creates a plain GtkEntry.
 */
static GtkWidget *
entry_new (KotoTaskEditor *editor)
{
  GtkWidget *entry;
  entry = gtk_entry_new ();
  g_signal_connect (entry, "changed", G_CALLBACK (on_entry_changed), editor);
  return entry;
}

/*
 * Editor setter function (an EditorSetFunc), that sets the value of a GtkEntry
 * to the string value of @prop.
 */
static void
entry_set_from_string (GtkWidget *widget, icalproperty *prop)
{
  const char *text = NULL;
  if (prop) {
    text = icalvalue_get_text (icalproperty_get_value (prop));
  }
  
  gtk_entry_set_text (GTK_ENTRY (widget), text ?: "");

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

#ifdef HAVE_SEXY
static void
open_url_entry (SexyIconEntry *entry, gint icon_pos, gint button, gpointer data)
{
  g_debug ("TODO: open URL %s", gtk_entry_get_text (GTK_ENTRY (entry)));
}
#endif

/*
 * Editor creation function (an EditorNew) , which creates an entry for URLs.
 */
static GtkWidget *
url_entry_new (KotoTaskEditor *editor)
{
  GtkWidget *entry;

#ifdef HAVE_SEXY
  GtkWidget *image;

  entry = sexy_icon_entry_new ();
  image = gtk_image_new_from_icon_name ("stock_internet", GTK_ICON_SIZE_MENU);

  sexy_icon_entry_set_icon (SEXY_ICON_ENTRY (entry),
                            SEXY_ICON_ENTRY_SECONDARY, GTK_IMAGE (image));
  /* TODO: set this depending on the content of the entry */
  sexy_icon_entry_set_icon_highlight (SEXY_ICON_ENTRY (entry),
                                      SEXY_ICON_ENTRY_SECONDARY, TRUE);

  g_signal_connect (entry, "icon-released", G_CALLBACK (open_url_entry), editor);
#else
  entry = gtk_entry_new ();
  g_signal_connect (entry, "changed", G_CALLBACK (on_entry_changed), editor);
#endif

  return entry;
}

static void
on_text_changed (GtkTextBuffer *buffer, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  GtkWidget *widget;
  GtkTextIter start, end;
  DisplayField *field;
  icalproperty *prop;
  const char *text;

  if (priv->ignore)
    return;

  widget = g_object_get_data (G_OBJECT (buffer), WIDGET_DATA);
  if (!widget) {
    g_warning (WIDGET_DATA " property not set");
    return;
  }

  field = g_object_get_data (G_OBJECT (widget), FIELD_DATA);
  prop = g_object_get_data (G_OBJECT (widget), PROP_DATA);

  if (!field) {
    g_warning (FIELD_DATA " property not set");
    return;
  }
  
  if (!prop) {
    prop = icalproperty_new (field->kind);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
  }

  gtk_text_buffer_get_start_iter (buffer, &start);
  gtk_text_buffer_get_end_iter (buffer, &end);

  text = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);

  /* NO means use the default value type. Ugly as sin. */
  icalproperty_set_value_from_string (prop, text, "NO");
  
  priv->dirty = TRUE;
  return;
}

static GtkWidget *
text_new (KotoTaskEditor *editor)
{
  GtkWidget *scrolled, *textview;
  GtkTextBuffer *buffer;

  scrolled = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled), GTK_SHADOW_IN);

  textview = gtk_text_view_new ();
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview), GTK_WRAP_WORD);
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  g_object_set_data (G_OBJECT (buffer), WIDGET_DATA, scrolled);
  gtk_widget_show (textview);
  gtk_container_add (GTK_CONTAINER (scrolled), textview);

  g_signal_connect (gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview)),
                    "changed", G_CALLBACK (on_text_changed),
                    editor);
  
  return scrolled;
}

static void
text_set_from_string (GtkWidget *widget, icalproperty *prop)
{
  const char *text = NULL;
  char *s;
  GtkWidget *textview;
  GtkTextBuffer *buffer;

  if (prop) {
    text = icalproperty_get_value_as_string (prop);
  }

  textview = gtk_bin_get_child (GTK_BIN (widget));
  if (!GTK_IS_TEXT_VIEW (textview)) {
    g_warning ("Child widget was not GtkTextView");
    return;
  }
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  if (text) {
    s = g_strcompress (text);
    gtk_text_buffer_set_text (buffer, s, -1);
    g_free (s);
  } else {
    gtk_text_buffer_set_text (buffer, "", -1);
  }
  
  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

static void
on_priority_changed (KotoPriorityCombo *combo, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  icalproperty *prop;
  int priority;

  if (priv->ignore)
    return;

  prop = g_object_get_data (G_OBJECT (combo), PROP_DATA);
  
  if (!prop) {
    prop = icalproperty_new (ICAL_PRIORITY_PROPERTY);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, prop);
  }

  priority = koto_priority_combo_get_priority (combo);
  if (priority != PRIORITY_NONE) {
    icalproperty_set_priority (prop, priority);
  } else {
    icalcomponent_remove_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, NULL);
  }

  priv->dirty = TRUE;
}

static GtkWidget *
priority_new (KotoTaskEditor *editor)
{
  GtkWidget *widget;
  widget = koto_priority_combo_new ();
  g_signal_connect (widget, "changed", G_CALLBACK (on_priority_changed), editor);
  return widget;
}

static void
priority_set (GtkWidget *widget, icalproperty *prop)
{
  int priority = 0;
  if (prop) {
    priority = icalproperty_get_priority (prop);
  }
  
  koto_priority_combo_set_priority (KOTO_PRIORITY_COMBO (widget), priority);

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

static void
on_date_notify (KotoDateCombo *combo, GParamSpec *spec, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  icalproperty *prop;
  GDate *date;
  icaltimetype due;

  if (priv->ignore)
    return;

  prop = g_object_get_data (G_OBJECT (combo), PROP_DATA);
  
  if (!prop) {
    prop = icalproperty_new (ICAL_DUE_PROPERTY);
    icalcomponent_add_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, prop);
  }

  date = koto_date_combo_get_date (combo);
  if (g_date_valid (date)) {
    /* This may be the fastest way from GDate to icaltimetype */
    due = icaltime_from_day_of_year (g_date_get_day_of_year (date),
                                     g_date_get_year (date));
    icalproperty_set_due (prop, due);
  } else {
    icalcomponent_remove_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, NULL);
  }

  priv->dirty = TRUE;
}

static GtkWidget *
date_new (KotoTaskEditor *editor)
{
  GtkWidget *widget;
  widget = koto_date_combo_new ();
  g_signal_connect (widget, "notify::date", G_CALLBACK (on_date_notify), editor);
  return widget;
}

static void
date_set (GtkWidget *widget, icalproperty *prop)
{
  GDate date;
  struct icaltimetype due;

  g_date_clear (&date, 1);
  
  if (prop) {
    due = icalproperty_get_due (prop);
    g_date_set_dmy (&date, due.day, due.month, due.year);
  }
  
  koto_date_combo_set_date (KOTO_DATE_COMBO (widget), &date);

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

static void
on_category_changed (KotoGroupCombo *combo, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  icalproperty *prop;
  KotoGroupType type;
  char *group;

  if (priv->ignore)
    return;

  prop = g_object_get_data (G_OBJECT (combo), PROP_DATA);
  
  type = koto_group_combo_get_active_type (combo);
  switch (type) {
  case GROUP_TYPE_NONE:
    icalcomponent_remove_property (priv->task->comp, prop);
    g_object_set_data (G_OBJECT (combo), PROP_DATA, NULL);
    break;
  case GROUP_TYPE_CATEGORY:
    /* Create the property if it didn't exist */
    if (!prop) {
      prop = icalproperty_new (ICAL_CATEGORIES_PROPERTY);
      icalcomponent_add_property (priv->task->comp, prop);
      g_object_set_data (G_OBJECT (combo), PROP_DATA, prop);
    }
    /* Now set its value */
    group = koto_group_combo_get_active_group (combo);
    icalproperty_set_categories (prop, group);
    g_free (group);
    break;
  case GROUP_TYPE_NEW:
    {
      GtkWidget *parent, *dialog, *entry;
      
      /* Magically get the top-level window */
      parent = gtk_widget_get_toplevel (GTK_WIDGET (combo));
      parent = GTK_WIDGET_TOPLEVEL (parent) ? parent : NULL;
      
      dialog = gtk_dialog_new_with_buttons (_("New Category"), GTK_WINDOW (parent),
                                            GTK_DIALOG_NO_SEPARATOR,
                                            GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                                            GTK_STOCK_OK, GTK_RESPONSE_OK,
                                            NULL);
      entry = gtk_entry_new ();
      gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), entry);
      gtk_widget_show_all (dialog);
      if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_OK) {
        GtkTreeIter iter;
        koto_group_store_add_new_group (priv->group_store, &iter, gtk_entry_get_text (GTK_ENTRY (entry)));
        gtk_combo_box_set_active_iter (GTK_COMBO_BOX (combo), &iter);
      } else {
        g_debug ("TODO: revert old group");
      }
      gtk_widget_destroy (dialog);
    }
    break;
  default:
    g_warning ("Unhandled group type %d", type);
    return;
  }

  priv->dirty = TRUE;
}

static GtkWidget *
category_new (KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  GtkWidget *combo;
  
  combo = koto_group_combo_new (priv->group_store);
  g_signal_connect (combo, "changed", G_CALLBACK (on_category_changed), editor);
  return combo;
}

static void
category_set (GtkWidget *widget, icalproperty *prop)
{
  char **categories;
  GtkTreeIter iter;

  if (prop) {
    categories = g_strsplit (icalproperty_get_categories (prop), ",", 0);
    koto_group_store_get_iter_for_group (KOTO_GROUP_STORE (gtk_combo_box_get_model (GTK_COMBO_BOX (widget))),
                                         categories[0], &iter);
    gtk_combo_box_set_active_iter (GTK_COMBO_BOX (widget), &iter);
    g_strfreev (categories);
  } else {
    gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
  }

  g_object_set_data (G_OBJECT (widget), PROP_DATA, prop);
}

/*
 * Called for every widget when a new icalcomponent is set.  If this widget has
 * an associated DisplayField set, then updates it with the value of the
 * relevant property.
 */
static void
update_field (GtkWidget *widget, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);
  DisplayField *field;
  icalproperty *prop;

  field = g_object_get_data (G_OBJECT (widget), FIELD_DATA);
  if (!field)
    return;

  prop = icalcomponent_get_first_property (priv->task->comp, field->kind);
  
  if (field->set_value) {
    field->set_value (widget, prop);
  } else {
    g_warning ("Field data for %s doesn't define set_value", field->display_name);
  }
}

/*
 * Called for every widget when a new KotoGroupStore is set.  If the widget is a
 * KotoGroupCombo, set the model property to the new store.
 */
static void
update_group_store (GtkWidget *widget, KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (editor);

  if (KOTO_IS_GROUP_COMBO (widget)) {
    g_object_set (widget, "model", priv->group_store, NULL);
  }
}

/*
 * GObject methods.
 */

static void
koto_task_editor_get_property (GObject *object, guint property_id,
                              GValue *value, GParamSpec *pspec)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_GROUPS:
    g_value_set_object (value, priv->group_store);
    break;
  case PROP_TASK:
    g_value_set_boxed (value, priv->task);
    break;
  case PROP_DIRTY:
    g_value_set_boolean (value, priv->dirty);
    break;
  default:
    if (property_id >= PROP_LAST && property_id < PROP_LAST + G_N_ELEMENTS (fields))
    {
      guint i;
      i = property_id - PROP_LAST;
      g_value_set_boolean (value,
          !gtk_widget_get_no_show_all (priv->field_widgets[i*2]));
    } else {
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
  }
}

static void
koto_task_editor_set_property (GObject *object, guint property_id,
                              const GValue *value, GParamSpec *pspec)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);

  switch (property_id) {
  case PROP_GROUPS:
    if (priv->group_store) {
      g_object_unref (priv->group_store);
      priv->group_store = NULL;
    }
    priv->group_store = KOTO_GROUP_STORE (g_value_dup_object (value));
    priv->ignore = TRUE;
    gtk_container_foreach (GTK_CONTAINER (object), (GtkCallback)update_group_store, object);
    priv->ignore = FALSE;
    break;
  case PROP_TASK:
    priv->task = g_value_dup_boxed (value);
    priv->dirty = FALSE;
    /* Now iterate over every widget, updating any fields with values from this icalcomponent. */
    priv->ignore = TRUE;
    gtk_container_foreach (GTK_CONTAINER (object), (GtkCallback)update_field, object);
    priv->ignore = FALSE;
    break;
  default:
    if (property_id >= PROP_LAST && property_id < PROP_LAST + G_N_ELEMENTS (fields))
    {
      guint i;
      gboolean visible = TRUE;
      visible = g_value_get_boolean (value);

      i = property_id - PROP_LAST;

      gtk_widget_set_no_show_all (priv->field_widgets[i*2], !visible);
      gtk_widget_set_no_show_all (priv->field_widgets[i*2+1], !visible);

      if (GTK_WIDGET_VISIBLE (priv->field_widgets[i*2]) && !visible)
        gtk_widget_hide (priv->field_widgets[i*2]);

      if (GTK_WIDGET_VISIBLE (priv->field_widgets[i*2+1]) && !visible)
        gtk_widget_hide (priv->field_widgets[i*2+1]);

      if (!GTK_WIDGET_VISIBLE (priv->field_widgets[i*2]) && visible)
        gtk_widget_show (priv->field_widgets[i*2]);

      if (!GTK_WIDGET_VISIBLE (priv->field_widgets[i*2+1]) && visible)
        gtk_widget_show (priv->field_widgets[i*2+1]);

    } else {
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
    }
  }
}

static void
koto_task_editor_dispose (GObject *object)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);
  
  if (priv->task) {
    koto_task_unref (priv->task);
    priv->task = NULL;
  }

  if (G_OBJECT_CLASS (koto_task_editor_parent_class)->dispose)
    G_OBJECT_CLASS (koto_task_editor_parent_class)->dispose (object);
}

static void
koto_task_editor_finalize (GObject *object)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (object);

  g_free (priv->field_widgets);

  G_OBJECT_CLASS (koto_task_editor_parent_class)->finalize (object);
}

static void
koto_task_editor_class_init (KotoTaskEditorClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  guint i;

  g_type_class_add_private (klass, sizeof (KotoTaskEditorPrivate));

  object_class->get_property = koto_task_editor_get_property;
  object_class->set_property = koto_task_editor_set_property;
  object_class->dispose = koto_task_editor_dispose;
  object_class->finalize = koto_task_editor_finalize;

  g_object_class_install_property (object_class, PROP_GROUPS,
                                   g_param_spec_object ("groups", "groups", NULL,
                                                        KOTO_TYPE_GROUP_STORE,
                                                        G_PARAM_READWRITE |
                                                        G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));
  g_object_class_install_property (object_class, PROP_TASK,
                                   g_param_spec_boxed ("task", "task", NULL,
                                                       KOTO_TYPE_TASK,
                                                       G_PARAM_READWRITE |
                                                       G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));
  g_object_class_install_property (object_class, PROP_DIRTY,
                                   g_param_spec_boolean ("dirty", "dirty", NULL,
                                                         FALSE,
                                                         G_PARAM_READABLE |
                                                         G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK));


  for (i = 0; i < G_N_ELEMENTS (fields); i++) {
    const DisplayField *field;

    field = &fields[i];

    g_object_class_install_property (object_class, PROP_LAST + i, 
        g_param_spec_boolean (field->property_name, field->property_name, NULL, TRUE,
          G_PARAM_READABLE | G_PARAM_WRITABLE |
          G_PARAM_STATIC_NAME | G_PARAM_STATIC_BLURB));
  }

}

static void
koto_task_editor_init (KotoTaskEditor *self)
{
  KotoTaskEditorPrivate *priv = GET_PRIVATE (self);
  int i;

  priv->field_widgets = g_malloc (sizeof (GtkWidget *) * G_N_ELEMENTS (fields) * 2);

  g_object_set (self,
                "homogeneous", FALSE,
                "row-spacing", 6,
                "column-spacing", 6,
                NULL);
  
  for (i = 0; i < G_N_ELEMENTS (fields); i++) {
    const DisplayField *field;
    GtkWidget *label, *widget;
    
    field = &fields[i];
    
    label = gtk_label_new_with_mnemonic (_(field->display_name));
    gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
    gtk_widget_show (label);
    gtk_table_attach (GTK_TABLE (self), label, 0, 1, i+1, i+2, GTK_FILL, GTK_FILL, 0, 0);

    widget = field->editor_new (self);
    g_object_set_data (G_OBJECT (widget), FIELD_DATA, (gpointer)field);
    gtk_widget_show (widget);

    if (GTK_IS_SCROLLED_WINDOW (widget)) {
      gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.0);
      gtk_table_attach_defaults (GTK_TABLE (self), widget, 1, 2, i+1, i+2);
    } else {
      gtk_table_attach (GTK_TABLE (self), widget, 1, 2, i+1, i+2, GTK_FILL, 0, 0, 0);
    }

    gtk_label_set_mnemonic_widget (GTK_LABEL (label), widget);

    priv->field_widgets[i*2] = label;
    priv->field_widgets[i*2+1] = widget;
  }
}


/*
 * Public methods.
 */

GtkWidget *
koto_task_editor_new (void)
{
  return g_object_new (KOTO_TYPE_TASK_EDITOR, NULL);
}

gboolean
koto_task_editor_is_dirty (KotoTaskEditor *editor)
{
  KotoTaskEditorPrivate *priv;

  g_return_val_if_fail (KOTO_IS_TASK_EDITOR (editor), FALSE);

  priv = GET_PRIVATE (editor);

  return priv->dirty;
}
